//
//  Viewer.java
//
//  Created by Thomas Jannaud on 09/08/08.
//  Copyright 2008 Jannaud's Technical Company. All rights reserved.
//

// _ Viewer_Oeil : il y a un plan P, un oeil O, et pour un point M de l'espace on note M' tel que O, M, M' soient alignés et M' € P
//			(en gros on trace (OM) et M' est le point en lequel (OM) coupe P.
//			Cette méthode est la plus réaliste concernant le rendu
// _ Viewer_Ortho : ne prend en compte qu'un plan P et elle réalise la projection orthogonale sur ce plan
// _ Une fois que l'on a l'image sur un plan, il suffit de prendre un repère (Oxy) et d'afficher notre projection.

class Point2D {
	public double x, y;
	Point2D(double a, double b) {
		x = a;
		y = b;
	}
}
class Point3D {
	public double x, y, z;
	Point3D(double a, double b, double c) {
		x = a;
		y = b;
		z = c;
	}
	
	public String toString() {
		return "x:" + x + " y:" + y + " z:" + z;
	}
}

public interface Viewer {
	Point2D projeter(Point3D p);
	int[] coordonnees(Point2D p);
	int[][] coordonnees(Point2D[] liste);
	boolean[] faces_visibles();
	void moveCamera(int dx, int dy);
}

class Viewer_Ortho implements Viewer {
	private double theta, phi, r, r2; // oeil en r2(cos phi * cos theta, cos phi * sin theta, sin phi) ; plan idem avec r
	private double a, b, c, d; // ax + by + cz = d
	
	Viewer_Ortho() {
		theta = 0.7;
		phi = 0.5;
		r = 2.0;
		calcule();
	}
	private void calcule() {
		a = Math.cos(phi) * Math.cos(theta);
		b = Math.cos(phi) * Math.sin(theta);
		c = Math.sin(phi);
		d = r;
	}
	
	public Point2D projeter(Point3D p) {
		double lambda = (d - a * p.x - b * p.y - c * p.z) / (a * a + b * b + c * c);
		Point3D proj = new Point3D(p.x + lambda * a, p.y + lambda * b, p.z + lambda * c);
		return new Point2D(-Math.sin(theta) * proj.x + Math.cos(theta) * proj.y, -Math.sin(phi) * (Math.cos(theta) * proj.x + Math.sin(theta) * proj.y) + Math.cos(phi) * proj.z);
	}
	public int[] coordonnees(Point2D p) {
		int[] retour = {(int)(p.x * 90.0) + 75, (int)(-p.y * 90.0) + 75};
		return retour;
	}
	public int[][] coordonnees(Point2D[] liste) {
		int n = liste.length;
		int[][] retour = new int[2][n];
		for(int i = 0; i < n; i++) {
			int[] a = coordonnees(liste[i]);
			retour[0][i] = a[0];
			retour[1][i] = a[1];
		}
		return retour;
	}
	
	public boolean[] faces_visibles() {
		boolean[] retour = new boolean[6];
		double[][] faces = {{0., 0., 1.}, {1., 0., 0.}, {0., -1., 0.}, {-1., 0., 0.}, {0., 1., 0.}, {0., 0., -1.}};
		for(int i = 0; i < 6; i++)
			retour[i] = ((a * faces[i][0] + b * faces[i][1] + c * faces[i][2]) > 0.0);
		return retour;
	}
	
	public void moveCamera(int dx, int dy) {
		if (Math.abs(Math.cos(phi)) < 0.1)
			theta += .1 * dx;
		else
			theta += .01 * dx / Math.cos(phi);
		phi += .01 * dy;
		calcule();
	}
}

class Viewer_Oeil implements Viewer {
	private double theta, phi, r, r2; // oeil en r2(cos phi * cos theta, cos phi * sin theta, sin phi) ; plan idem avec r
	private double a, b, c, d; // ax + by + cz = d
	private Point3D oeil;
	
	Viewer_Oeil() {
		theta = 0.7;
		phi = 0.5;
		r = 2.0;
		r2 = 3.0;
		oeil = new Point3D(r2 * Math.cos(phi) * Math.cos(theta), r2 * Math.cos(phi) * Math.sin(theta), r2 * Math.sin(phi));
		calcule();
	}
	private void calcule() {
		a = Math.cos(phi) * Math.cos(theta);
		b = Math.cos(phi) * Math.sin(theta);
		c = Math.sin(phi);
		d = r;
		oeil.x = r2 * Math.cos(phi) * Math.cos(theta);
		oeil.y = r2 * Math.cos(phi) * Math.sin(theta);
		oeil.z = r2 * Math.sin(phi);
	}

	public Point2D projeter(Point3D p) {
		double lambda = (d - a * oeil.x - b * oeil.y - c * oeil.z) / (a * (p.x - oeil.x) + b * (p.y - oeil.y) + c * (p.z - oeil.z));
		Point3D proj = new Point3D(oeil.x + lambda * (p.x - oeil.x), oeil.y + lambda * (p.y - oeil.y), oeil.z + lambda * (p.z - oeil.z));
		return new Point2D(-Math.sin(theta) * proj.x + Math.cos(theta) * proj.y, -Math.sin(phi) * (Math.cos(theta) * proj.x + Math.sin(theta) * proj.y) + Math.cos(phi) * proj.z);
	}
	public int[] coordonnees(Point2D p) {
		int[] retour = {(int)(p.x * 200.0) + 65, (int)(-p.y * 200.0) + 70};
		return retour;
	}
	public int[][] coordonnees(Point2D[] liste) {
		int n = liste.length;
		int[][] retour = new int[2][n];
		for(int i = 0; i < n; i++) {
			int[] a = coordonnees(liste[i]);
			retour[0][i] = a[0];
			retour[1][i] = a[1];
		}
		return retour;
	}
	
	public boolean[] faces_visibles() {
		boolean[] retour = new boolean[6];
		double[][] faces = {{0., 0., 1.}, {1., 0., 0.}, {0., -1., 0.}, {-1., 0., 0.}, {0., 1., 0.}, {0., 0., -1.}};
		for(int i = 0; i < 6; i++)
			retour[i] = ((a * faces[i][0] + b * faces[i][1] + c * faces[i][2]) > 0.15);
		return retour;
	}
	
	public void moveCamera(int dx, int dy) {
		if (Math.abs(Math.cos(phi)) < 0.1)
			theta += .1 * dx;
		else
			theta += .01 * dx / Math.cos(phi);
		phi += .01 * dy;
		calcule();
	}
}